/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2016  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#ifndef LASERDATAVIEWERHEADERFILE
#define LASERDATAVIEWERHEADERFILE

#include <QMainWindow>

#include "laserfile.h"

class QPlainTextEdit;
class QPushButton;
class QWidget;
class QToolButton;
class QDragEnterEvent;
class QDropEvent;

class TGraphWidget;

class TLaserDataViewer : public QMainWindow
{
Q_OBJECT

public:

// CREATORS
	TLaserDataViewer(const QWidget *parent=0, Qt::WindowFlags flags=0);
	~TLaserDataViewer(void);

// ACCESSORS

// MANIPULATORS

signals:

private slots:	
	void Import_Data(void);
	void Save_As(void);
	
	void Copy_Dev_Normal(void);
	void Copy_Dev_Invert(void);
	void Copy_All_Data(void);
	
	void Section_1_Toggled(bool);
	void Section_2_Toggled(bool);
	void Section_3_Toggled(bool);

protected:
	virtual void dragEnterEvent(QDragEnterEvent *DragEnterEvent);
	virtual void dropEvent(QDropEvent *DropEvent);

private:
	QPlainTextEdit						*d_data_text;
	QPushButton							*d_saveas_button;
	QPushButton							*d_copy_inverted_button;
	QPushButton							*d_copy_normal_button;
	QPushButton							*d_copy_all_button;
	QToolButton							*d_section_1_button;
	QToolButton							*d_section_2_button;
	QToolButton							*d_section_3_button;
	TGraphWidget						*d_graph_display_widget;
	
	TLaserFile							d_laser_file;
	
	int Get_Active_Section(void) const;
	
	void Load_Data(const QString &file_name);
	void Update_Display(void);
	void Update_Section_Names(void);
	
	void Clear_Display(void);
	void Add_Display_Text(const QString &text);
	void Add_Display_Text(const QStringList &text_lines);
	
	void Clear_Graph(void);
	void Enable_Options(bool state);


// NOT IMPLEMENTED
	TLaserDataViewer(const TLaserDataViewer&);
	TLaserDataViewer& operator=(const TLaserDataViewer&);
};

#endif
