/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2016  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QString>
#include <QFile>
#include <QTextStream>
#include <QRegExp>
#include <QDir>
#include <assert.h>

#include "laserfile.h"


TLaserFile::TLaserFile(void)
{
}

TLaserFile::~TLaserFile(void)
{
}

QString TLaserFile::File_Name_Short(void) const
{
	int									index;
	
	index = d_file_name.lastIndexOf('/');
	return d_file_name.mid(index + 1);
}

int TLaserFile::Data_Count(const int section) const
{
	switch(section)
	{
		case 0:
			return static_cast<int>(d_data_section_1.size());
			break;
			
		case 1:
			return static_cast<int>(d_data_section_2.size());
			break;
			
		case 2:
			return static_cast<int>(d_data_section_3.size());
			break;
	}
	
	return 0;
}

std::vector<TVector2> TLaserFile::Data(
	const int							section,
	const int							index) const
{
	std::vector<TVector2>				average;
	std::vector<TVector2>::iterator		average_iter;
	std::vector<TVector2>::const_iterator iter;
	std::vector<std::vector<TVector2> >::const_iterator array_iter;
	bool								init;
	TVector2							pnt;

	if(section == 0)
	{
		if(index < 0)
		{
			if(d_data_section_1.size() > 0)
			{
				// return average of all data
				init = true;
				
				average.clear();
				
				for(array_iter = d_data_section_1.begin();array_iter != d_data_section_1.end();++array_iter)
				{
					if(init)
					{
						average = (*array_iter);
						init = false;
					}
					else
					{
						for(iter = (*array_iter).begin(),average_iter = average.begin();
							iter != (*array_iter).end();
							++iter,++average_iter)
						{
							(*average_iter) += (*iter);
						}
					}
				}
				
				for(average_iter = average.begin();average_iter != average.end();++average_iter)
				{
					(*average_iter) /= static_cast<double>(d_data_section_1.size());
				}
			}

			return average;
		}
		
		assert(static_cast<unsigned int>(index) < d_data_section_1.size());
		
		return d_data_section_1[index];
	}
	else if(section == 1)
	{
		if(index < 0)
		{
			if(d_data_section_2.size() > 0)
			{
				// return average of all data
				init = true;
				
				average.clear();
				
				for(array_iter = d_data_section_2.begin();array_iter != d_data_section_2.end();++array_iter)
				{
					if(init)
					{
						average = (*array_iter);
						init = false;
					}
					else
					{
						for(iter = (*array_iter).begin(),average_iter = average.begin();
							iter != (*array_iter).end();
							++iter,++average_iter)
						{
							(*average_iter) += (*iter);
						}
					}
				}
				
				for(average_iter = average.begin();average_iter != average.end();++average_iter)
				{
					(*average_iter) /= static_cast<double>(d_data_section_2.size());
				}
			}
			
			return average;
		}
		
		assert(static_cast<unsigned int>(index) < d_data_section_2.size());
		
		return d_data_section_2[index];
	}
	else if(section == 2)
	{
		if(index < 0)
		{
			if(d_data_section_3.size() > 0)
			{
				// return average of all data
				init = true;
				
				average.clear();
				
				for(array_iter = d_data_section_3.begin();array_iter != d_data_section_3.end();++array_iter)
				{
					if(init)
					{
						average = (*array_iter);
						init = false;
					}
					else
					{
						for(iter = (*array_iter).begin(),average_iter = average.begin();
							iter != (*array_iter).end();
							++iter,++average_iter)
						{
							(*average_iter) += (*iter);
						}
					}
				}
				
				for(average_iter = average.begin();average_iter != average.end();++average_iter)
				{
					(*average_iter) /= static_cast<double>(d_data_section_3.size());
				}
			}
			
			return average;
		}
		
		assert(static_cast<unsigned int>(index) < d_data_section_3.size());
		
		return d_data_section_3[index];
	}

	return average; // empty array
}

QString TLaserFile::File_Type_Description(void) const
{
	QString								description;
	
	switch(d_file_type)
	{
		case TLaserFile::TYPE_UNKNOWN:
			description = QStringLiteral("Unknown file type or file not loaded.");
			break;
			
		case TLaserFile::TYPE_RENISHAW_RTL:
			description = QStringLiteral("Renishaw ML10/XL80 linear data (RTL).");
			break;
			
		case TLaserFile::TYPE_RENISHAW_RTA:
			description = QStringLiteral("Renishaw ML10/XL80 angular data (RTA).");
			break;
			
		case TLaserFile::TYPE_RENISHAW_ST:
			description = QStringLiteral("Renishaw ML10/XL80 straightness data (STx).");
			break;
			
		case TLaserFile::TYPE_API_SCALE:
			description = QStringLiteral("API XD scale data (POS).");
			break;
			
		case TLaserFile::TYPE_API_STRAIGHTNESS:
			description = QStringLiteral("API XD straightness data (STR).");
			break;
			
		case TLaserFile::TYPE_API_ANGULAR:
			description = QStringLiteral("API XD angular data (ANG).");
			break;
			
	}
	
	return description;
}

QString TLaserFile::Section_Name(
	const int							section) const
{
	QString								name;
	
	if(section == 0)
	{
		switch(d_file_type)
		{
			case TLaserFile::TYPE_UNKNOWN:
				name = QStringLiteral("X");
				break;
				
			case TLaserFile::TYPE_RENISHAW_RTL:
			case TLaserFile::TYPE_API_SCALE:
				name = QStringLiteral("L1");
				break;
				
			case TLaserFile::TYPE_RENISHAW_RTA:
				name = QStringLiteral("A1");
				break;
				
			case TLaserFile::TYPE_RENISHAW_ST:
				name = QStringLiteral("S1.");
				break;

			case TLaserFile::TYPE_API_STRAIGHTNESS:
				name = QStringLiteral("XX");
				break;
				
			case TLaserFile::TYPE_API_ANGULAR:
				name = QStringLiteral("RA");
				break;
				
		}
	}
	else if(section == 1)
	{
		switch(d_file_type)
		{
			case TLaserFile::TYPE_UNKNOWN:
			case TLaserFile::TYPE_RENISHAW_RTL:
			case TLaserFile::TYPE_API_SCALE:
			case TLaserFile::TYPE_RENISHAW_RTA:
			case TLaserFile::TYPE_RENISHAW_ST:
				name = QStringLiteral("X");
				break;
				
			case TLaserFile::TYPE_API_STRAIGHTNESS:
				name = QStringLiteral("YY");
				break;
				
			case TLaserFile::TYPE_API_ANGULAR:
				name = QStringLiteral("RB");
				break;
				
		}
	}
	else if(section == 2)
	{
		switch(d_file_type)
		{
			case TLaserFile::TYPE_UNKNOWN:
			case TLaserFile::TYPE_RENISHAW_RTL:
			case TLaserFile::TYPE_API_SCALE:
			case TLaserFile::TYPE_RENISHAW_RTA:
			case TLaserFile::TYPE_RENISHAW_ST:
			case TLaserFile::TYPE_API_STRAIGHTNESS:
				name = QStringLiteral("X");
				break;
				
			case TLaserFile::TYPE_API_ANGULAR:
				name = QStringLiteral("RC");
				break;
				
		}
	}
	else
	{
		name = QStringLiteral("X");
	}
	
	return name;
}

void TLaserFile::Reset(void)
{
	d_data_section_1.clear();
	d_data_section_2.clear();
	d_data_section_3.clear();
	
	d_environment.air_temperature = 20;
	d_environment.air_pressure = 100;
	d_environment.air_humidity = 50;
	d_environment.mat_temperature = 20;
	
	d_file_name = QString();
	d_file_type = TLaserFile::TYPE_UNKNOWN;
}

bool TLaserFile::Import_File(
	const QString						&file_name)
{
	QFile								file;
	QByteArray							file_data;
	TLaserFile::TLaserFileType			file_type;
	
	file.setFileName(file_name);
	
	if(!file.open(QIODevice::ReadOnly))
	{
		d_last_error = QString("Cannot open file '%1' for reading.").arg(QDir::toNativeSeparators(file_name));
		return false;
	}
	
	file_data = file.readAll();
	file.close();
	
	if(file_data.size() < 100)	// too small to have anything
	{
		d_last_error = QString("Input file '%1' is too small to be a laser file.").arg(QDir::toNativeSeparators(file_name));
		return false;
	}
	
	file_type = this->Determine_File_Type(file_name,file_data);
	
	if(file_type == TLaserFile::TYPE_UNKNOWN)
	{
		d_last_error = QString("Input file '%1' doesn't appear to be a known laser file type.").arg(QDir::toNativeSeparators(file_name));
		return false;
	}
	
	if(file_type == TLaserFile::TYPE_RENISHAW_RTL ||
	   file_type == TLaserFile::TYPE_RENISHAW_RTA ||
	   file_type == TLaserFile::TYPE_RENISHAW_ST)
	{
		if(!this->Import_Renishaw_File(file_data))
		{
			return false;
		}
	}
	else if(file_type == TLaserFile::TYPE_API_SCALE)
	{
		if(!this->Import_API_Scale_File(file_data))
		{
			return false;
		}
	}
	else if(file_type == TLaserFile::TYPE_API_STRAIGHTNESS)
	{
		if(!this->Import_API_Straightness_File(file_data))
		{
			return false;
		}
	}
	else if(file_type == TLaserFile::TYPE_API_ANGULAR)
	{
		if(!this->Import_API_Angular_File(file_data))
		{
			return false;
		}
	}
	else
	{
		d_last_error = QString("Input file '%1' doesn't appear to be a known laser file type.").arg(QDir::toNativeSeparators(file_name));
		return false;
	}

	d_file_name = file_name;
	d_file_type = file_type;

	return true;
}

TLaserFile::TLaserFileType TLaserFile::Determine_File_Type(
	const QString						&file_name,
	const QByteArray					&file_data) const
{
	Q_UNUSED(file_data);
	
	QString								extension;
	int									index;
	
	// currently only considering the file extension
	// scanning the file data may be added in the future to solve collisions
	index = file_name.lastIndexOf('.');
	
	if(index < 0)
	{
		return TLaserFile::TYPE_UNKNOWN;
	}
	
	extension = file_name.mid(index + 1);
	
	if(extension.compare(QStringLiteral("RTL"),Qt::CaseInsensitive) == 0)
	{
		return TLaserFile::TYPE_RENISHAW_RTL;
	}
	else if(extension.compare(QStringLiteral("RTA"),Qt::CaseInsensitive) == 0)
	{
		return TLaserFile::TYPE_RENISHAW_RTA;
	}
	else if(extension.startsWith(QStringLiteral("STR"),Qt::CaseInsensitive))
	{
		return TLaserFile::TYPE_API_STRAIGHTNESS;
	}
	else if(extension.startsWith(QStringLiteral("ST"),Qt::CaseInsensitive))
	{
		return TLaserFile::TYPE_RENISHAW_ST;
	}
	else if(extension.startsWith(QStringLiteral("POS"),Qt::CaseInsensitive))
	{
		return TLaserFile::TYPE_API_SCALE;
	}
	else if(extension.startsWith(QStringLiteral("ANG"),Qt::CaseInsensitive))
	{
		return TLaserFile::TYPE_API_ANGULAR;
	}
	
	return TLaserFile::TYPE_UNKNOWN;
}

bool TLaserFile::Import_Renishaw_File(
	const QByteArray					&file_data)
{
	QString								text;
	bool								inside_target_section(false);
	bool								found_target_data(false);
	bool								inside_deviation_section(false);
	bool								found_deviation_data(false);
	bool								inside_environment_section(false);
	bool								material_sensor_1_valid(false);
	bool								material_sensor_2_valid(false);
	bool								material_sensor_3_valid(false);
	bool								conversion_state;
	QStringList							text_values;
	std::vector<double>					target_values;
	double								dval;
	double								mat1_temperature(0);
	double								mat2_temperature(0);
	double								mat3_temperature(0);
	int									item_count;
	int									target_count;
	int									cntr;
	TLaserFile::TRenDevData				deviation_item;
	std::vector<TLaserFile::TRenDevData>	deviation_data;
	
	d_has_environmental_data = false;
	target_count = 0;
	
	QTextStream text_stream(file_data,QIODevice::ReadOnly);
	
	while(!text_stream.atEnd())
	{
		text = text_stream.readLine().simplified();
		
		if(text.length() == 0)
		{
			continue;
		}
		
		if(text.startsWith(QStringLiteral("TARGET DATA::"),Qt::CaseInsensitive))
		{
			inside_target_section = true;
			inside_deviation_section = false;
			inside_environment_section = false;
			
			continue;
		}
		else if(text.startsWith(QStringLiteral("DEVIATIONS::"),Qt::CaseInsensitive))
		{
			inside_deviation_section = true;
			inside_target_section = false;
			inside_environment_section = false;
			
			continue;
		}
		else if(text.startsWith(QStringLiteral("ENVIRONMENT::"),Qt::CaseInsensitive))
		{
			inside_environment_section = true;
			inside_target_section = false;
			inside_deviation_section = false;
			
			d_has_environmental_data = true;
			
			continue;
		}
		
		// special case searches
		// these datafiles vary quite a bit so the section headers may not exist
		// maybe leftover from the DOS days ...
	
		if(inside_target_section == false &&
		   inside_deviation_section == false &&
		   inside_environment_section == false)
		{
			if(text.startsWith(QStringLiteral("Targets :"),Qt::CaseInsensitive) || text.startsWith(QStringLiteral("Targets:"),Qt::CaseInsensitive))
			{
				inside_target_section = true;
				found_target_data = true;
				
				inside_deviation_section = false;
				inside_environment_section = false;
				
				continue;
			}
			else if(text.startsWith(QStringLiteral("Run Target Data"),Qt::CaseInsensitive))
			{
				inside_deviation_section = true;
				found_deviation_data = true;
				
				inside_target_section = false;
				inside_environment_section = false;
				
				continue;
			}
		}

		// look for target count parameter.  this parameter must exist
		// some older files do not identify the sections so this cannot be relied upon
		if(inside_deviation_section == false && inside_environment_section == false)
		{
			if(text.startsWith(QStringLiteral("Target-count"),Qt::CaseInsensitive))
			{
				// text may be 'Target-count: xx' or 'Target-count=xx' - bleh!
				if(text.contains(':'))
				{
					text_values = text.split(':');
				}
				else if(text.contains('='))
				{
					text_values = text.split('=');
				}
				else
				{
					d_last_error = QString("Unexpected text in TARGET DATA:: section: '%1'").arg(text);
					return false;
				}
				
				if(text_values.size() != 2)
				{
					d_last_error = QString("Unexpected text in TARGET DATA:: section: '%1'").arg(text);
					return false;
				}
				
				dval = text_values[1].toDouble(&conversion_state);
				
				if(!conversion_state)
				{
					d_last_error = QString("Unexpected text in TARGET DATA:: section: '%1'").arg(text);
					return false;
				}

				if(dval < 1)
				{
					d_last_error = QString("Unexpected text in TARGET DATA:: section: '%1'").arg(text);
					return false;
				}
				
				target_count = static_cast<int>(dval + 0.5);	// round value just in case
			}
		}
		
		// targets
		if(inside_target_section == true && found_target_data == true)
		{
			// read data lines as target values
			
			text_values = text.split(' ',QString::SkipEmptyParts);
			
			item_count = text_values.size();
			
			for(cntr = 0;cntr < item_count;cntr++)
			{
				dval = text_values[cntr].toDouble(&conversion_state);
				
				if(conversion_state)
				{
					target_values.push_back(dval);
				}
				else
				{
					inside_target_section = false;
					break;
				}
			}
		}
		else if(inside_target_section == true && found_target_data == false)
		{
			if(text.startsWith(QStringLiteral("Targets:"),Qt::CaseInsensitive) || text.startsWith(QStringLiteral("Targets :"),Qt::CaseInsensitive))
			{
				found_target_data = true;
			}
		}
		else if(text.startsWith(QStringLiteral("EOF::"),Qt::CaseInsensitive))
		{
			inside_target_section = false;
			inside_deviation_section = false;
			inside_environment_section = false;
		}

		// deviations
		if(inside_deviation_section == true && found_deviation_data == true)
		{
			text_values = text.split(' ',QString::SkipEmptyParts);
			
			item_count = text_values.size();
			
			if(item_count > 2)
			{
				deviation_item.run_id = text_values[0].toInt(&conversion_state);
				
				if(!conversion_state)
				{
					inside_deviation_section = false;
					break;
				}
				
				deviation_item.target_id = text_values[1].toInt(&conversion_state);
				
				if(!conversion_state)
				{
					inside_deviation_section = false;
					break;
				}
				
				deviation_item.deviation = text_values[2].toDouble(&conversion_state);
				
				if(!conversion_state)
				{
					inside_deviation_section = false;
					break;
				}
				
				deviation_data.push_back(deviation_item);
			}
		}
		else if(inside_deviation_section == true && found_deviation_data == false)
		{
			if(text.startsWith(QStringLiteral("Run Target Data"),Qt::CaseInsensitive))
			{
				found_deviation_data = true;
			}
		}
		else if(text.startsWith(QStringLiteral("EOF::"),Qt::CaseInsensitive))
		{
			inside_target_section = false;
			inside_deviation_section = false;
			inside_environment_section = false;
		}

		
		// environment
		if(inside_environment_section)
		{
			// all values must be valid
			if(text.startsWith(QStringLiteral("Air temp"),Qt::CaseInsensitive))
			{
				text_values = text.split(':');
				
				if(text_values.size() != 2)
				{
					d_last_error = QString("Unexpected text in ENVIRONMENT:: section: '%1'").arg(text);
					return false;
				}
				
				text = text_values[1];
				text_values = text.split(' ',QString::SkipEmptyParts);
				
				if(text_values.size() > 1)
				{
					dval = text_values[0].toDouble(&conversion_state);
					
					if(!conversion_state)
					{
						d_last_error = QString("Unexpected text in ENVIRONMENT:: section: '%1'").arg(text);
						return false;
					}
					
					dval += text_values[1].toDouble(&conversion_state);
					
					if(!conversion_state)
					{
						d_last_error = QString("Unexpected text in ENVIRONMENT:: section: '%1'").arg(text);
						return false;
					}
					
					d_environment.air_temperature = dval / 2.0;
				}
				else
				{
					d_last_error = QString("Unexpected text in ENVIRONMENT:: section: '%1'").arg(text);
					return false;
				}
			}
			else if(text.startsWith(QStringLiteral("Air press"),Qt::CaseInsensitive))
			{
				text_values = text.split(':');
				
				if(text_values.size() != 2)
				{
					d_last_error = QString("Unexpected text in ENVIRONMENT:: section: '%1'").arg(text);
					return false;
				}
				
				text = text_values[1];
				text_values = text.split(' ',QString::SkipEmptyParts);
				
				if(text_values.size() > 1)
				{
					dval = text_values[0].toDouble(&conversion_state);
					
					if(!conversion_state)
					{
						d_last_error = QString("Unexpected text in ENVIRONMENT:: section: '%1'").arg(text);
						return false;
					}
					
					dval += text_values[1].toDouble(&conversion_state);
					
					if(!conversion_state)
					{
						d_last_error = QString("Unexpected text in ENVIRONMENT:: section: '%1'").arg(text);
						return false;
					}
					
					d_environment.air_pressure = dval / 2.0;
				}
				else
				{
					d_last_error = QString("Unexpected text in ENVIRONMENT:: section: '%1'").arg(text);
					return false;
				}
			}
			else if(text.startsWith(QStringLiteral("Air humid"),Qt::CaseInsensitive))
			{
				text_values = text.split(':');
				
				if(text_values.size() != 2)
				{
					d_last_error = QString("Unexpected text in ENVIRONMENT:: section: '%1'").arg(text);
					return false;
				}
				
				text = text_values[1];
				text_values = text.split(' ',QString::SkipEmptyParts);
				
				if(text_values.size() > 1)
				{
					dval = text_values[0].toDouble(&conversion_state);
					
					if(!conversion_state)
					{
						d_last_error = QString("Unexpected text in ENVIRONMENT:: section: '%1'").arg(text);
						return false;
					}
					
					dval += text_values[1].toDouble(&conversion_state);
					
					if(!conversion_state)
					{
						d_last_error = QString("Unexpected text in ENVIRONMENT:: section: '%1'").arg(text);
						return false;
					}
					
					d_environment.air_humidity = dval / 2.0;
				}
				else
				{
					d_last_error = QString("Unexpected text in ENVIRONMENT:: section: '%1'").arg(text);
					return false;
				}
			}
			else if(text.startsWith(QStringLiteral("Mat temp 1"),Qt::CaseInsensitive))
			{
				text_values = text.split(':');
				
				if(text_values.size() != 2)
				{
					d_last_error = QString("Unexpected text in ENVIRONMENT:: section: '%1'").arg(text);
					return false;
				}
				
				text = text_values[1];
				text_values = text.split(' ',QString::SkipEmptyParts);
				
				if(text_values.size() > 1)
				{
					dval = text_values[0].toDouble(&conversion_state);
					
					if(!conversion_state)
					{
						d_last_error = QString("Unexpected text in ENVIRONMENT:: section: '%1'").arg(text);
						return false;
					}
					
					dval += text_values[1].toDouble(&conversion_state);
					
					if(!conversion_state)
					{
						d_last_error = QString("Unexpected text in ENVIRONMENT:: section: '%1'").arg(text);
						return false;
					}
					
					if(dval > 0)
					{
						material_sensor_1_valid = true;
						mat1_temperature = dval / 2;
					}
				}
				else
				{
					d_last_error = QString("Unexpected text in ENVIRONMENT:: section: '%1'").arg(text);
					return false;
				}
			}
			else if(text.startsWith(QStringLiteral("Mat temp 2"),Qt::CaseInsensitive))
			{
				text_values = text.split(':');
				
				if(text_values.size() != 2)
				{
					d_last_error = QString("Unexpected text in ENVIRONMENT:: section: '%1'").arg(text);
					return false;
				}
				
				text = text_values[1];
				text_values = text.split(' ',QString::SkipEmptyParts);
				
				if(text_values.size() > 1)
				{
					dval = text_values[0].toDouble(&conversion_state);
					
					if(!conversion_state)
					{
						d_last_error = QString("Unexpected text in ENVIRONMENT:: section: '%1'").arg(text);
						return false;
					}
					
					dval += text_values[1].toDouble(&conversion_state);
					
					if(!conversion_state)
					{
						d_last_error = QString("Unexpected text in ENVIRONMENT:: section: '%1'").arg(text);
						return false;
					}
					
					if(dval > 0)
					{
						material_sensor_2_valid = true;
						mat2_temperature = dval / 2;
					}
				}
				else
				{
					d_last_error = QString("Unexpected text in ENVIRONMENT:: section: '%1'").arg(text);
					return false;
				}
			}
			else if(text.startsWith(QStringLiteral("Mat temp 3"),Qt::CaseInsensitive))
			{
				text_values = text.split(':');
				
				if(text_values.size() != 2)
				{
					d_last_error = QString("Unexpected text in ENVIRONMENT:: section: '%1'").arg(text);
					return false;
				}
				
				text = text_values[1];
				text_values = text.split(' ',QString::SkipEmptyParts);
				
				if(text_values.size() > 1)
				{
					dval = text_values[0].toDouble(&conversion_state);
					
					if(!conversion_state)
					{
						d_last_error = QString("Unexpected text in ENVIRONMENT:: section: '%1'").arg(text);
						return false;
					}
					
					dval += text_values[1].toDouble(&conversion_state);
					
					if(!conversion_state)
					{
						d_last_error = QString("Unexpected text in ENVIRONMENT:: section: '%1'").arg(text);
						return false;
					}
					
					if(dval > 0)
					{
						material_sensor_3_valid = true;
						mat3_temperature = dval / 2;
					}
				}
				else
				{
					d_last_error = QString("Unexpected text in ENVIRONMENT:: section: '%1'").arg(text);
					return false;
				}
			}
			else if(text.startsWith(QStringLiteral("Final Data"),Qt::CaseInsensitive))
			{
				inside_environment_section = false;
			}
		}
		else if(text.startsWith(QStringLiteral("EOF::"),Qt::CaseInsensitive))
		{
			inside_target_section = false;
			inside_deviation_section = false;
			inside_environment_section = false;
		}
	}
	
	if(target_count < 1)
	{
		d_last_error = QString("Parameter 'Target-count' not found in file.  Cannot process data.");
		return false;
	}
	
	if(static_cast<int>(target_values.size()) > target_count)
	{
		target_values.resize(target_count);
	}
	
	deviation_data = this->Sort_Deviation_Data(deviation_data);
	this->Write_Data_Arrays(target_values,deviation_data);
	
	// find average of valid sensors
	if(material_sensor_1_valid && material_sensor_2_valid && material_sensor_3_valid)
	{
		d_environment.mat_temperature = (mat1_temperature + mat2_temperature + mat3_temperature)/ 3.0;
	}
	else if(material_sensor_1_valid && material_sensor_2_valid)
	{
		d_environment.mat_temperature = (mat1_temperature + mat2_temperature)/ 2.0;
	}
	else if(material_sensor_1_valid && material_sensor_3_valid)
	{
		d_environment.mat_temperature = (mat1_temperature + mat3_temperature)/ 2.0;
	}
	else if(material_sensor_2_valid && material_sensor_3_valid)
	{
		d_environment.mat_temperature = (mat2_temperature + mat3_temperature)/ 2.0;
	}
	else if(material_sensor_1_valid)
	{
		d_environment.mat_temperature = mat1_temperature;
	}
	else if(material_sensor_2_valid)
	{
		d_environment.mat_temperature = mat2_temperature;
	}
	else if(material_sensor_3_valid)
	{
		d_environment.mat_temperature = mat3_temperature;
	}

	return true;
}

std::vector<TLaserFile::TRenDevData> TLaserFile::Sort_Deviation_Data(
	const std::vector<TLaserFile::TRenDevData> &deviation_data) const
{
	std::vector<TLaserFile::TRenDevData>::const_iterator dev_iter;
	std::vector<TLaserFile::TRenDevData> run_data;
	std::vector<TLaserFile::TRenDevData> sorted_data;
	bool								init(true);
	int									run;

	// data sorted to match targets.  data is bidirectional in data files
	
	for(dev_iter = deviation_data.begin();dev_iter != deviation_data.end();++dev_iter)
	{
		if(init)
		{
			run = (*dev_iter).run_id;
			init = false;
		}
		
		if(run != (*dev_iter).run_id)
		{
			run = (*dev_iter).run_id;
			
			sorted_data.insert(sorted_data.end(),run_data.begin(),run_data.end());
			
			run_data.clear();
		}
		
		if(run_data.size())
		{
			// add to beginning or end based on value of 'Target'
			
			if((*dev_iter).target_id > (*(run_data.end() - 1)).target_id)
			{
				run_data.push_back(*dev_iter);
			}
			else
			{
				run_data.insert(run_data.begin(),(*dev_iter));
			}
		}
		else
		{
			run_data.push_back(*dev_iter);
		}
	}
	
	// check for unprocessed data
	if(run_data.size())
	{
		sorted_data.insert(sorted_data.end(),run_data.begin(),run_data.end());
	}
	
	return sorted_data;
}

void TLaserFile::Write_Data_Arrays(
	const std::vector<double>			&target_data,
	const std::vector<TLaserFile::TRenDevData> &deviation_data)
{
	std::vector<TLaserFile::TRenDevData> sorted_deviation_data;
	std::vector<TLaserFile::TRenDevData>::const_iterator dev_iter;
	std::vector<double>::const_iterator	target_iter;
	std::vector<TVector2>				data_array;
	std::vector<TVector2>::iterator		data_iter;
	bool								init(true);
	int									run;
	TVector2							pnt;
	
	d_data_section_1.clear();
	d_data_section_2.clear();
	d_data_section_3.clear();
	
	for(dev_iter = deviation_data.begin();dev_iter != deviation_data.end();++dev_iter)
	{
		if(init)
		{
			run = (*dev_iter).run_id;
			init = false;
		}
		
		if(run != (*dev_iter).run_id)
		{
			run = (*dev_iter).run_id;
			
			if(!this->Is_Null_Array(target_data.size(),data_array))
			{
				for(target_iter = target_data.begin(),data_iter = data_array.begin();
					target_iter != target_data.end();
					++target_iter,++data_iter)
				{
					(*data_iter).x = (*target_iter);
				}
				
				d_data_section_1.push_back(data_array);
			}
			
			data_array.clear();
		}
		
		pnt.y = (*dev_iter).deviation;
		
		data_array.push_back(pnt);
	}
	
	// check for unprocessed data
	if(!this->Is_Null_Array(target_data.size(),data_array))
	{
		for(target_iter = target_data.begin(),data_iter = data_array.begin();
			target_iter != target_data.end();
			++target_iter,++data_iter)
		{
			(*data_iter).x = (*target_iter);
		}
		
		d_data_section_1.push_back(data_array);
	}
}

bool TLaserFile::Is_Null_Array(
	const unsigned int					target_size,
	const std::vector<TVector2>			&array_data) const
{
	std::vector<TVector2>::const_iterator iter;
	const double						ZERO_EPSILON(0.0000001);

	if(array_data.size() != target_size)
	{
		return true;
	}
	
	for(iter = array_data.begin();iter != array_data.end();++iter)
	{
		if(fabs((*iter).y) > ZERO_EPSILON)
		{
			return false;
		}
	}
	
	return true;
}

bool TLaserFile::Import_API_Scale_File(
	const QByteArray					&file_data)
{
	QString								text;
	bool								conversion_state;
	QStringList							values_text;
	int									item_count;
	bool								inside_measurement(false);
	bool								inside_environment(false);
	double								initial_value;
	TLaserFile::TMeasurement			measurement;
	std::vector<TLaserFile::TMeasurement> measurements;
	double								last_measurement_position;
	double								unit_conversion(0);
	QRegExp								linear_pattern("^([-]?[0-9.]{1,12})[ ]([-]?[0-9.]{1,16})[ ]([-]?[0-9.]{1,10})");
	const double						MINIMUM_DISTANCE(0.5);

	measurement.type = TLaserFile::SCALE;
	measurement.direction = TLaserFile::FORWARD;
	measurement.err_scale = 0;
	measurement.err_xx = 0;
	measurement.err_yy = 0;
	measurement.err_ra=0;
	measurement.err_rb=0;
	measurement.err_rc=0;
	
	initial_value = 0;
	last_measurement_position = 0;
	d_has_environmental_data = false;

	QTextStream text_stream(file_data,QIODevice::ReadOnly);
	
	while(!text_stream.atEnd())
	{
		text = text_stream.readLine();
		text = text.simplified();
		
		if(text.length() == 0)
		{
			inside_environment = false;
			inside_measurement = false;
			
			continue;
		}
		
		// IniPos:	-475
		// EndPos:	475			don't need this?
		
		if(text.startsWith(QStringLiteral("IniPos:"),Qt::CaseInsensitive))
		{
			if(!unit_conversion)
			{
				d_last_error = QStringLiteral("Scale file measurement units unknown");
				return false;
			}
			
			values_text = text.split(':',QString::SkipEmptyParts);
			item_count = values_text.size();
			
			if(item_count == 2)	// Description,Value
			{
				initial_value = values_text[1].toDouble(&conversion_state);
				initial_value *= unit_conversion;
				
				last_measurement_position = initial_value;
				
				if(!conversion_state)
				{
					d_last_error = QString("Scale file unexpected text: '%1'").arg(text);
					return false;
				}
			}
			else
			{
				d_last_error = QString("Scale file unexpected text: '%1'").arg(text);
				return false;
			}
			
			continue;
		}
		else if(text.startsWith("Unit:") || text.startsWith("Units:"))	// both 'Unit:' and 'Units:' are used
		{
			values_text = text.split(':',QString::SkipEmptyParts);
			item_count = values_text.size();
			
			if(item_count == 2)	// Description,Value
			{
				text = values_text[1].simplified();
				
				if(text.startsWith("mm",Qt::CaseInsensitive))
				{
					unit_conversion = 1.0;
				}
				else if (text.startsWith("inch",Qt::CaseInsensitive))
				{
					unit_conversion = 25.4;
				}
				else
				{
					d_last_error = QString("Scale file unknown units: '%1'").arg(text);
					return false;
				}
				
			}
			else
			{
				d_last_error = QString("Scale file unexpected text: '%1'").arg(text);
				return false;
			}
			
			continue;
		}
		else if(text.startsWith(QStringLiteral("<Environment>"),Qt::CaseInsensitive))
		{
			inside_environment = true;
			inside_measurement = false;
			
			d_has_environmental_data = true;

			
			continue;
		}
		else if(text.contains(QStringLiteral("Position"),Qt::CaseInsensitive) &&
				text.contains(QStringLiteral("AirTemp"),Qt::CaseInsensitive) &&
				text.contains(QStringLiteral("MatTemp"),Qt::CaseInsensitive))
		{
			inside_environment = false;
			inside_measurement = true;
			
			continue;
		}

		if(inside_measurement)
		{
			//    Position	      LasRead	  LinearError	Status	AirTemp	AirPress	MatTemp	MatTemp2	MatTemp3
			//	-475.00000	      0.00010	      0.00010	     0	20.00	732.24	24.10
			// Only first three is considered.  Second value is actual position, third is error
			
			if(!(linear_pattern.indexIn(text) < 0) && linear_pattern.captureCount() == 3)
			{
				if(unit_conversion == 0)
				{
					d_last_error = QString("Scale file units not found prior to data: '%1'").arg(text);
					return false;
				}
				
				measurement.position = linear_pattern.cap(1).toDouble();
				measurement.position *= unit_conversion;
				
				measurement.err_scale = linear_pattern.cap(3).toDouble() * unit_conversion;
				
				if(measurements.size())
				{
					if(measurement.direction == TLaserFile::FORWARD)
					{
						if(!((measurement.position - last_measurement_position + MINIMUM_DISTANCE) > 0) ||
						   fabs(measurement.position - last_measurement_position) < MINIMUM_DISTANCE)
						{
							measurement.direction = TLaserFile::BACKWARD;
						}
					}
					else if(measurement.direction == TLaserFile::BACKWARD)
					{
						if(!((measurement.position - last_measurement_position + MINIMUM_DISTANCE) < 0) ||
						   fabs(measurement.position - last_measurement_position) < MINIMUM_DISTANCE)
						{
							measurement.direction = TLaserFile::FORWARD;
						}
					}
				}
				
				measurements.push_back(measurement);
				last_measurement_position = measurement.position;
				
			}
			else if(inside_measurement)
			{
				inside_measurement = false;
			}
		}
		else if(inside_environment)
		{
			if(text.startsWith(QStringLiteral("Average Air Temperature"),Qt::CaseInsensitive))
			{
				values_text = text.split(':',QString::SkipEmptyParts);
				item_count = values_text.size();
				
				if(item_count == 2)	// Description,Value
				{
					d_environment.air_temperature = values_text[1].toDouble(&conversion_state);

					if(!conversion_state)
					{
						d_last_error = QString("Scale file unexpected text: '%1'").arg(text);
						return false;
					}
				}
				else
				{
					d_last_error = QString("Scale file unexpected text: '%1'").arg(text);
					return false;
				}
			}
			else if(text.startsWith(QStringLiteral("Average Material"),Qt::CaseInsensitive))
			{
				values_text = text.split(':',QString::SkipEmptyParts);
				item_count = values_text.size();
				
				if(item_count == 2)	// Description,Value
				{
					d_environment.mat_temperature = values_text[1].toDouble(&conversion_state);
					
					if(!conversion_state)
					{
						d_last_error = QString("Scale file unexpected text: '%1'").arg(text);
						return false;
					}
				}
				else
				{
					d_last_error = QString("Scale file unexpected text: '%1'").arg(text);
					return false;
				}
			}
			else if(text.startsWith(QStringLiteral("Average Air Pressure"),Qt::CaseInsensitive))
			{
				values_text = text.split(':',QString::SkipEmptyParts);
				item_count = values_text.size();
				
				if(item_count == 2)	// Description,Value
				{
					d_environment.air_pressure = values_text[1].toDouble(&conversion_state);
					d_environment.air_pressure /= 7.500616827042;	// convert from mmHg to kPa
					
					if(!conversion_state)
					{
						d_last_error = QString("Scale file unexpected text: '%1'").arg(text);
						return false;
					}
				}
				else
				{
					d_last_error = QString("Scale file unexpected text: '%1'").arg(text);
					return false;
				}
			}
		}
	}
	
	measurements = this->Sort_Measurement_Data(measurements);
	this->Write_Data_Arrays_Scale(measurements);

	return true;
}

bool TLaserFile::Import_API_Straightness_File(
	const QByteArray					&file_data)
{
	QString								text;
	bool								conversion_state;
	QStringList							values_text;
	int									item_count;
	bool								inside_measurement(false);
	double								initial_value(0.0);
	TLaserFile::TMeasurement			measurement;
	std::vector<TLaserFile::TMeasurement> measurements;
	double								last_measurement_position;
	double								unit_conversion(0);
	QRegExp								straightness_pattern("^([-]?[0-9.]{1,12})[ ]([-]?[0-9.]{1,16})[ ]([-]?[0-9.]{1,16})");
	const double						MINIMUM_DISTANCE(0.5);
	
	measurement.type = TLaserFile::STRAIGHTNESS;
	measurement.direction = TLaserFile::FORWARD;
	measurement.err_scale = 0;
	measurement.err_xx = 0;
	measurement.err_yy = 0;
	measurement.err_ra=0;
	measurement.err_rb=0;
	measurement.err_rc=0;
	last_measurement_position = 0;
	
	d_has_environmental_data = false;

	QTextStream text_stream(file_data,QIODevice::ReadOnly);
	
	while(!text_stream.atEnd())
	{
		text = text_stream.readLine();
		text = text.simplified();
		
		if(text.length() == 0)
		{
			inside_measurement = false;
			
			continue;
		}
		
		// IniPos:	-475
		// EndPos:	475			don't need this?
		
		if(text.startsWith("IniPos:"))
		{
			if(!unit_conversion)
			{
				d_last_error = QStringLiteral("Straightness file measurement units unknown");
				return false;
			}
			
			values_text = text.split(':',QString::SkipEmptyParts);
			item_count = values_text.size();
			
			if(item_count == 2)	// Description,Value
			{
				initial_value = values_text[1].toDouble(&conversion_state);
				initial_value *= unit_conversion;
				
				last_measurement_position = initial_value;
				
				if(!conversion_state)
				{
					d_last_error = QString("Straightness file unexpected text: '%1'").arg(text);
					return false;
				}
			}
			else
			{
				d_last_error = QString("Straightness file unexpected text: '%1'").arg(text);
				return false;
			}
			
			continue;
		}
		else if(text.startsWith("Unit:") || text.startsWith("Units:"))	// both 'Unit:' and 'Units:' are used
		{
			values_text = text.split(':',QString::SkipEmptyParts);
			item_count = values_text.size();
			
			if(item_count == 2)	// Description,Value
			{
				text = values_text[1].simplified();
				
				if(text.startsWith("mm",Qt::CaseInsensitive))
				{
					unit_conversion = 1.0;
				}
				else if (text.startsWith("inch",Qt::CaseInsensitive))
				{
					unit_conversion = 25.4;
				}
				else
				{
					d_last_error = QString("Straightness file unexpected text: '%1'").arg(text);
					return false;
				}
			}
			else
			{
				d_last_error = QString("Straightness file unknown units: '%1'").arg(text);
				return false;
			}
			
			continue;
		}
		else if(text.contains(QStringLiteral("Position"),Qt::CaseInsensitive) &&
				text.contains(QStringLiteral("XX"),Qt::CaseInsensitive) &&
				text.contains(QStringLiteral("YY"),Qt::CaseInsensitive))
		{
			inside_measurement = true;
			
			continue;
		}
		
		if(inside_measurement)
		{
			// Position	            XX	            YY	Status
			// -475.00000	       0.00000	       0.00000	     0
			// Position, horizontal straightness, vertical straightness

			if(!(straightness_pattern.indexIn(text) < 0) && straightness_pattern.captureCount() == 3)
			{
				if(unit_conversion == 0)
				{
					d_last_error = QString("Straightness file units not found prior to data: '%1'").arg(text);
					return false;
				}
				
				measurement.position = straightness_pattern.cap(1).toDouble();
				measurement.position *= unit_conversion;
	//			measurement.position -= initial_value;
				
				measurement.err_xx = straightness_pattern.cap(2).toDouble() * unit_conversion;
				measurement.err_yy = straightness_pattern.cap(3).toDouble() * unit_conversion;
				
				if(measurements.size())
				{
					if(measurement.direction == TLaserFile::FORWARD)
					{
						if(!((measurement.position - last_measurement_position + MINIMUM_DISTANCE) > 0) ||
						   fabs(measurement.position - last_measurement_position) < MINIMUM_DISTANCE)
						{
							measurement.direction = TLaserFile::BACKWARD;
						}
					}
					else if(measurement.direction == TLaserFile::BACKWARD)
					{
						if(!((measurement.position - last_measurement_position + MINIMUM_DISTANCE) < 0) ||
						   fabs(measurement.position - last_measurement_position) < MINIMUM_DISTANCE)
						{
							measurement.direction = TLaserFile::FORWARD;
						}
					}
				}
				
				measurements.push_back(measurement);
				last_measurement_position = measurement.position;
				
			}
			else
			{
				inside_measurement = false;
				break;	// no more measurement data
			}
		}
	}
	
	measurements = this->Sort_Measurement_Data(measurements);
	this->Write_Data_Arrays_Straightness(measurements);
	
	return true;
}

bool TLaserFile::Import_API_Angular_File(
	const QByteArray					&file_data)
{
	QString								text;
	bool								conversion_state;
	QStringList							values_text;
	int									item_count;
	bool								inside_measurement(false);
	double								initial_value(0.0);
	TLaserFile::TMeasurement			measurement;
	std::vector<TLaserFile::TMeasurement> measurements;
	double								last_measurement_position;
	double								unit_conversion(0);
	double								angle_conversion(0);
	QRegExp								angular_pattern("^([-]?[0-9.]{1,12})[ ]([-]?[0-9.]{1,16})[ ]([-]?[0-9.]{1,16})[ ]([-]?[0-9.]{1,16})");
	const double						MINIMUM_DISTANCE(0.5);
	
	measurement.type = TLaserFile::ANGULAR;
	measurement.direction = TLaserFile::FORWARD;
	measurement.err_scale = 0;
	measurement.err_xx = 0;
	measurement.err_yy = 0;
	measurement.err_ra=0;
	measurement.err_rb=0;
	measurement.err_rc=0;
	last_measurement_position = 0;
	
	d_has_environmental_data = false;

	QTextStream text_stream(file_data,QIODevice::ReadOnly);
	
	while(!text_stream.atEnd())
	{
		text = text_stream.readLine();
		text = text.simplified();
		
		if(text.length() == 0)
		{
			inside_measurement = false;
			
			continue;
		}
		
		// IniPos:	-475
		// EndPos:	475			don't need this?
		
		if(text.startsWith("IniPos:"))
		{
			if(!unit_conversion)
			{
				d_last_error = QStringLiteral("Angular file measurement units unknown");
				return false;
			}
			
			values_text = text.split(':',QString::SkipEmptyParts);
			item_count = values_text.size();
			
			if(item_count == 2)	// Description,Value
			{
				initial_value = values_text[1].toDouble(&conversion_state);
				initial_value *= unit_conversion;
				
				last_measurement_position = initial_value;
				
				if(!conversion_state)
				{
					d_last_error = QString("Angular file unexpected text: '%1'").arg(text);
					return false;
				}
			}
			else
			{
				d_last_error = QString("Angular file unexpected text: '%1'").arg(text);
				return false;
			}
			
			continue;
		}
		else if(text.startsWith("Unit:") || text.startsWith("Units:"))	// both 'Unit:' and 'Units:' are used
		{
			values_text = text.split(':',QString::SkipEmptyParts);
			item_count = values_text.size();
			
			if(item_count == 2)	// Description,Value
			{
				text = values_text[1].simplified();
				
				if(text.startsWith("mm",Qt::CaseInsensitive))
				{
					unit_conversion = 1.0;
				}
				else if (text.startsWith("inch",Qt::CaseInsensitive))
				{
					unit_conversion = 25.4;
				}
				else
				{
					d_last_error = QString("Angular file unexpected text: '%1'").arg(text);
					return false;
				}
				
				if(text.contains("arc sec",Qt::CaseInsensitive))
				{
					angle_conversion = 0.0048481368110764;	// conversion from arc sec to mm/m
				}
				else if(text.contains("in/ft",Qt::CaseInsensitive))
				{
					angle_conversion = 83.3333333333333333;	// conversion from inch/foot to to mm/m
				}
				else if(text.contains("mm/m",Qt::CaseInsensitive))
				{
					angle_conversion = 1.0;
				}
				else
				{
					d_last_error = QString("Angular file unknown units: '%1'").arg(text);
					return false;
				}
			}
			else
			{
				d_last_error = QString("Angular file unexpected text: '%1'").arg(text);
				return false;
			}
			
			continue;
		}
		else if(text.contains(QStringLiteral("Position"),Qt::CaseInsensitive) &&
				text.contains(QStringLiteral("AngA"),Qt::CaseInsensitive) &&
				text.contains(QStringLiteral("AngB"),Qt::CaseInsensitive) &&
				text.contains(QStringLiteral("AngC"),Qt::CaseInsensitive))
		{
			inside_measurement = true;
			
			continue;
		}
		
		if(inside_measurement)
		{
			//  Position	      AngA	      AngB	      AngC	LevelMst	LevelRef	   Status
			// -475.00000	      0.00	      0.00	     -0.00	     36.29	      4.22	        0
			// Position, angle a, angle b, angle c
			
			if(!(angular_pattern.indexIn(text) < 0) && angular_pattern.captureCount() == 4)
			{
				if(unit_conversion == 0)
				{
					d_last_error = QString("Angular file units not found prior to data: '%1'").arg(text);
					return false;
				}
				
				measurement.position = angular_pattern.cap(1).toDouble();
				measurement.position *= unit_conversion;
	//			measurement.position -= initial_value;
				
				measurement.err_ra = angular_pattern.cap(2).toDouble() * angle_conversion;
				measurement.err_rb = angular_pattern.cap(3).toDouble() * angle_conversion;
				measurement.err_rc = angular_pattern.cap(4).toDouble() * angle_conversion;
				
				if(measurements.size())
				{
					if(measurement.direction == TLaserFile::FORWARD)
					{
						if(!((measurement.position - last_measurement_position + MINIMUM_DISTANCE) > 0) ||
						   fabs(measurement.position - last_measurement_position) < MINIMUM_DISTANCE)
						{
							measurement.direction = TLaserFile::BACKWARD;
						}
					}
					else if(measurement.direction == TLaserFile::BACKWARD)
					{
						if(!((measurement.position - last_measurement_position + MINIMUM_DISTANCE) < 0) ||
						   fabs(measurement.position - last_measurement_position) < MINIMUM_DISTANCE)
						{
							measurement.direction = TLaserFile::FORWARD;
						}
					}
				}
				
				measurements.push_back(measurement);
				last_measurement_position = measurement.position;
				
			}
			else
			{
				inside_measurement = false;
				break;	// no more measurement data
			}
		}
	}
	
	measurements = this->Sort_Measurement_Data(measurements);
	this->Write_Data_Arrays_Angular(measurements);
	
	return true;
}

std::vector<TLaserFile::TMeasurement> TLaserFile::Sort_Measurement_Data(
	 const std::vector<TLaserFile::TMeasurement> &measurement_data) const
{
	std::vector<TLaserFile::TMeasurement>::const_iterator meas_iter;
	std::vector<TLaserFile::TMeasurement> run_data;
	std::vector<TLaserFile::TMeasurement> sorted_data;
	TLaserFile::TDirection				direction;
	bool								init(true);
	
	// data sorted to match targets.  data is bidirectional in data files
	
	for(meas_iter = measurement_data.begin();meas_iter != measurement_data.end();++meas_iter)
	{
		if(init)
		{
			direction = (*meas_iter).direction;
			init = false;
		}
		
		if(direction != (*meas_iter).direction)
		{
			direction = (*meas_iter).direction;
			
			sorted_data.insert(sorted_data.end(),run_data.begin(),run_data.end());
			
			run_data.clear();
		}
		
		if(run_data.size())
		{
			// add to beginning or end based on value of 'direction'
			
			if((*meas_iter).direction == TLaserFile::FORWARD)
			{
				run_data.push_back(*meas_iter);
			}
			else
			{
				run_data.insert(run_data.begin(),(*meas_iter));
			}
		}
		else
		{
			run_data.push_back(*meas_iter);
		}
	}
	
	// check for unprocessed data
	if(run_data.size())
	{
		sorted_data.insert(sorted_data.end(),run_data.begin(),run_data.end());
	}
	
	return sorted_data;
}

void TLaserFile::Write_Data_Arrays_Scale(const std::vector<
	TLaserFile::TMeasurement>			&measurement_data)
{
	std::vector<TLaserFile::TMeasurement>::const_iterator meas_iter;
	std::vector<TVector2>				data_array;
	TVector2							pnt;
	TLaserFile::TDirection				direction;
	bool								init(true);
	
	d_data_section_1.clear();
	d_data_section_2.clear();
	d_data_section_3.clear();

	for(meas_iter = measurement_data.begin();meas_iter != measurement_data.end();++meas_iter)
	{
		if(init)
		{
			direction = (*meas_iter).direction;
			init = false;
		}

		if((*meas_iter).direction != direction)
		{
			direction = (*meas_iter).direction;
			
			d_data_section_1.push_back(data_array);
			data_array.clear();
		}
		
		pnt.x = (*meas_iter).position;
		pnt.y = (*meas_iter).err_scale;
		
		data_array.push_back(pnt);
	}
	
	// check for unprocessed data
	if(data_array.size())
	{
		d_data_section_1.push_back(data_array);
	}
}


void TLaserFile::Write_Data_Arrays_Straightness(const std::vector<
	TLaserFile::TMeasurement>			&measurement_data)
{
	std::vector<TLaserFile::TMeasurement>::const_iterator meas_iter;
	std::vector<TVector2>				data_array_s1;
	std::vector<TVector2>				data_array_s2;
	TVector2							pnt;
	TLaserFile::TDirection				direction;
	bool								init(true);
	
	d_data_section_1.clear();
	d_data_section_2.clear();
	d_data_section_3.clear();

	for(meas_iter = measurement_data.begin();meas_iter != measurement_data.end();++meas_iter)
	{
		if(init)
		{
			direction = (*meas_iter).direction;
			init = false;
		}
		
		if((*meas_iter).direction != direction)
		{
			direction = (*meas_iter).direction;
			
			d_data_section_1.push_back(data_array_s1);
			d_data_section_2.push_back(data_array_s2);
			
			data_array_s1.clear();
			data_array_s2.clear();
		}
		
		pnt.x = (*meas_iter).position;
		pnt.y = (*meas_iter).err_xx;
		
		data_array_s1.push_back(pnt);
		
		pnt.y = (*meas_iter).err_yy;
		data_array_s2.push_back(pnt);
	}
	
	// check for unprocessed data
	if(data_array_s1.size())
	{
		d_data_section_1.push_back(data_array_s1);
	}
	
	if(data_array_s2.size())
	{
		d_data_section_2.push_back(data_array_s2);
	}
}

void TLaserFile::Write_Data_Arrays_Angular(const std::vector<
	TLaserFile::TMeasurement>			&measurement_data)
{
	std::vector<TLaserFile::TMeasurement>::const_iterator meas_iter;
	std::vector<TVector2>				data_array_s1;
	std::vector<TVector2>				data_array_s2;
	std::vector<TVector2>				data_array_s3;
	TVector2							pnt;
	TLaserFile::TDirection				direction;
	bool								init(true);
	
	d_data_section_1.clear();
	d_data_section_2.clear();
	d_data_section_3.clear();

	for(meas_iter = measurement_data.begin();meas_iter != measurement_data.end();++meas_iter)
	{
		if(init)
		{
			direction = (*meas_iter).direction;
			init = false;
		}
		
		if((*meas_iter).direction != direction)
		{
			direction = (*meas_iter).direction;
			
			d_data_section_1.push_back(data_array_s1);
			d_data_section_2.push_back(data_array_s2);
			d_data_section_3.push_back(data_array_s3);
			
			data_array_s1.clear();
			data_array_s2.clear();
			data_array_s3.clear();
		}
		
		pnt.x = (*meas_iter).position;
		pnt.y = (*meas_iter).err_ra;
		
		data_array_s1.push_back(pnt);
		
		pnt.y = (*meas_iter).err_rb;
		data_array_s2.push_back(pnt);
		
		pnt.y = (*meas_iter).err_rc;
		data_array_s3.push_back(pnt);
	}
	
	// check for unprocessed data
	if(data_array_s1.size())
	{
		d_data_section_1.push_back(data_array_s1);
	}
	
	if(data_array_s2.size())
	{
		d_data_section_2.push_back(data_array_s2);
	}
	
	if(data_array_s3.size())
	{
		d_data_section_3.push_back(data_array_s3);
	}
}

