/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2016  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#ifndef LASERFILEHEADERFILE
#define LASERFILEHEADERFILE

#include <QByteArray>
#include <vector>

#include "../../core/vector2.h"

class TLaserFile
{
public:
	
struct TLaserDataEnvironment
{
	double								air_temperature;
	double								air_pressure;
	double								air_humidity;
	double								mat_temperature;
};

	
// CREATORS
	TLaserFile(void);
	~TLaserFile(void);

// ACCESSORS
	QString Get_Last_Error(void) const {return d_last_error;}
	
	QString File_Name(void) const {return d_file_name;}
	QString File_Name_Short(void) const;
	
	int Data_Count(const int section) const;
	std::vector<TVector2> Data(const int section,const int index) const;	// -1 for index returns average
	
	bool Has_Envirionmental_Data(void) const {return d_has_environmental_data;}
	TLaserDataEnvironment Environmental_Data(void) const {return d_environment;}
	
	QString File_Type_Description(void) const;
	QString Section_Name(const int section) const;
	
// MANIPULATORS
	void Reset(void);
	bool Import_File(const QString &file_name);
	
private:

enum TLaserFileType
{
	TYPE_UNKNOWN = 0,
	TYPE_RENISHAW_RTL,
	TYPE_RENISHAW_RTA,
	TYPE_RENISHAW_ST,
	TYPE_API_SCALE,
	TYPE_API_STRAIGHTNESS,
	TYPE_API_ANGULAR
};
	
struct TRenDevData
{
	int									run_id;
	int									target_id;
	double								deviation;
};

enum TMeasurementType
{
	SCALE=0,
	STRAIGHTNESS,
	ANGULAR,
	SCALE_STRAIGHTNESS_ANGULAR,
	COMBINED
};
	
enum TDirection
{
	FORWARD=0,
	BACKWARD
};

struct TMeasurement
{
	TMeasurementType					type;
	double								position;
	TDirection							direction;
	double								err_scale;
	double								err_xx;
	double								err_yy;
	double								err_ra;
	double								err_rb;
	double								err_rc;
};

	TLaserDataEnvironment				d_environment;
	std::vector<std::vector<TVector2> >	d_data_section_1;
	std::vector<std::vector<TVector2> >	d_data_section_2;
	std::vector<std::vector<TVector2> >	d_data_section_3;
	QString								d_file_name;
	TLaserFileType						d_file_type;
	
	bool								d_has_environmental_data;
	
	mutable QString						d_last_error;
	
	TLaserFileType Determine_File_Type(const QString &file_name,const QByteArray &file_data) const;
	
	bool Import_Renishaw_File(const QByteArray &file_data);
	std::vector<TLaserFile::TRenDevData> Sort_Deviation_Data(const std::vector<TLaserFile::TRenDevData> &deviation_data) const;
	void Write_Data_Arrays(const std::vector<double> &target_data,const std::vector<TLaserFile::TRenDevData> &deviation_data);
	bool Is_Null_Array(const unsigned int target_size,const std::vector<TVector2> &array_data) const;
	
	bool Import_API_Scale_File(const QByteArray &file_data);
	bool Import_API_Straightness_File(const QByteArray &file_data);
	bool Import_API_Angular_File(const QByteArray &file_data);
	std::vector<TLaserFile::TMeasurement> Sort_Measurement_Data(const std::vector<TLaserFile::TMeasurement> &measurement_data) const;
	void Write_Data_Arrays_Scale(const std::vector<TLaserFile::TMeasurement> &measurement_data);
	void Write_Data_Arrays_Straightness(const std::vector<TLaserFile::TMeasurement> &measurement_data);
	void Write_Data_Arrays_Angular(const std::vector<TLaserFile::TMeasurement> &measurement_data);

// NOT IMPLEMENTED
	TLaserFile(const TLaserFile&);
	TLaserFile& operator=(const TLaserFile&);
};

#endif